# -*- coding: utf-8 -*-

API_VERSION = 'v3'
API_USER_AGENT = 'Etu Nexus API Python Client' + '-' + API_VERSION

__version__ = '0.6.7'
__url__ = 'http://www.etunexus.com/'
__author_email__ = 'indichen@etusolution.com'
__author__ = 'Etu Corporation'
__license__ = 'GPL'